#  Copyright (C) 2021-2022 YADRO.
#  All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import os
import time

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import units

from cinder import context as cinder_context
from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder import utils
from cinder.volume import configuration
from cinder.volume import driver
from cinder.volume.drivers.san import san
from cinder.volume.drivers.yadro.tatlin_client import InitTatlinClient
from cinder.volume.drivers.yadro.tatlin_exception import TatlinAPIException
from cinder.volume.drivers.yadro.tatlin_utils import TatlinVolumeConnections
from cinder.volume import qos_specs
from cinder.volume import volume_types
from cinder.volume import volume_utils
from cinder.volume.volume_utils import brick_get_connector_properties


LOG = logging.getLogger(__name__)

tatlin_opts = [
    cfg.StrOpt('pool_name',
               default='',
               help='storage pool name'),
    cfg.PortOpt('api_port',
                default=443,
                help='Port to use to access the Tatlin API'),
    cfg.StrOpt('export_ports',
               default='',
               help='Ports to export Tatlin resource through'),
    cfg.StrOpt('host_group',
               default='',
               help='Tatlin host group name'),
    cfg.IntOpt('max_resource_count',
               default=500,
               help='Max resource count allowed for Tatlin'),
    cfg.IntOpt('pool_max_resource_count',
               default=250,
               help='Max resource count allowed for single pool'),
    cfg.IntOpt('tat_api_retry_count',
               default=10,
               help='Number of retry on Tatlin API'),
    cfg.StrOpt('auth_method',
               default='CHAP',
               help='Authentication method for iSCSI (CHAP)'),
    cfg.StrOpt('lba_format',
               default='512e',
               help='LBA Format for new volume'),
    cfg.IntOpt('wait_retry_count',
               default=15,
               help='Number of checks for a lengthy operation to finish'),
    cfg.IntOpt('wait_interval',
               default=30,
               help='Wait number of seconds before re-checking'),
]

CONF = cfg.CONF
CONF.register_opts(tatlin_opts, group=configuration.SHARED_CONF_GROUP)


class TatlinCommonVolumeDriver(driver.VolumeDriver, object):

    def __init__(self, *args, **kwargs):
        super(TatlinCommonVolumeDriver, self).__init__(*args, **kwargs)
        self._ip = None
        self._port = 443
        self._user = None
        self._password = None
        self._pool_name = None
        self._pool_id = None
        self.configuration.append_config_values(san.san_opts)
        self.configuration.append_config_values(tatlin_opts)
        self._auth_method = 'CHAP'
        self._chap_username = ''
        self._chap_password = ''
        self.backend_name = None
        self.DRIVER_VOLUME_TYPE = None
        self._export_ports = None
        self._host_group = None
        self.verify = None
        self.DEFAULT_FILTER_FUNCTION = None
        self.DEFAULT_GOODNESS_FUNCTION = None
        self._use_multipath = True
        self._enforce_multipath = False
        self._lba_format = '512e'
        self._ssl_cert_path = None
        self._max_pool_resource_count = 250

    def do_setup(self, context):
        """Initial driver setup"""
        required_config = ['san_ip',
                           'san_login',
                           'san_password',
                           'pool_name',
                           'host_group']
        for attr in required_config:
            if not getattr(self.configuration, attr, None):
                message = (_('config option %s is not set.') % attr)
                raise exception.InvalidInput(message=message)

        self._ip = self.configuration.san_ip
        self._user = self.configuration.san_login
        self._password = self.configuration.san_password
        self._port = self.configuration.api_port
        self._pool_name = self.configuration.pool_name
        self._export_ports = self.configuration.export_ports
        self._host_group = self.configuration.host_group
        self._auth_method = self.configuration.auth_method
        self._chap_username = self.configuration.chap_username
        self._chap_password = self.configuration.chap_password
        self._wait_interval = self.configuration.wait_interval
        self._wait_retry_count = self.configuration.wait_retry_count

        self._ssl_cert_path = (self.configuration.
                               safe_get('driver_ssl_cert_path') or None)

        self.verify = (self.configuration.
                       safe_get('driver_ssl_cert_verify') or False)

        if self.verify and self._ssl_cert_path:
            self.verify = self._ssl_cert_path

        LOG.info('Tatlin driver version: %s', self.VERSION)

        self.tatlin_api = self._get_tatlin_client()
        self.ctx = context
        self.MAX_ALLOWED_RESOURCES = self.configuration.max_resource_count
        self._max_pool_resource_count = \
            self.configuration.pool_max_resource_count
        self.DEFAULT_FILTER_FUNCTION = \
            'capabilities.pool_resource_count < ' +\
            str(self._max_pool_resource_count) +\
            ' and capabilities.overall_resource_count < ' +\
            str(self.MAX_ALLOWED_RESOURCES)
        self.DEFAULT_GOODNESS_FUNCTION = '100 - capabilities.utilization'
        self._use_multipath = \
            (self.configuration.safe_get(
                'use_multipath_for_image_xfer') or False)
        self._enforce_multipath = \
            (self.configuration.safe_get(
                'enforce_multipath_for_image_xfer') or False)
        self._lba_format = self.configuration.lba_format
        self._wait_interval = self.configuration.wait_interval
        self._wait_retry_count = self.configuration.wait_retry_count
        self._connections = TatlinVolumeConnections(
            os.path.join(CONF.state_path,
                         'tatlin-volume-connections'))

    def check_for_setup_error(self):
        pass

    @volume_utils.trace
    def create_volume(self, volume):
        """Entry point for create new volume"""

        if not self.pool_id:
            raise exception.VolumeBackendAPIException(
                message='Wrong Tatlin pool configuration')

        pool_res_count, cluster_res_count = \
            self.tatlin_api.get_resource_count(self.pool_id)
        LOG.debug('Current pool %(pool)s has %(pool_res)s res.'
                  'Whole cluster has %(cluster_res)s',
                  {'pool': self.pool_id,
                   'pool_res': pool_res_count,
                   'cluster_res': cluster_res_count})

        self._stats['pool_resource_count'] = pool_res_count
        self._stats['overall_resource_count'] = cluster_res_count

        if pool_res_count > 255:
            message = _('TatlinVolumeDriver create volume failed. '
                        'Too many resources per pool created')
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

        if cluster_res_count + 1 > self.MAX_ALLOWED_RESOURCES:
            message = _('TatlinVolumeDriver create volume failed. '
                        'Too many resources per cluster created')
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

        LOG.debug('Create volume %(vol_id)s started',
                  {'vol_id': volume.name_id})
        self._create_volume_storage(volume)
        LOG.debug('Create volume %s finished', volume.name_id)

    def _create_volume_storage(self, volume):
        """Create a volume with a specific name in Tatlin"""
        size = volume.size * units.Gi
        vol_type = 'snapshot' if 'snapshot_volume' in volume.metadata \
            else 'volume'
        name = 'cinder-%s-%s' % (vol_type, volume.name_id)
        LOG.debug('Creating Tatlin resource %(name)s '
                  'with %(size)s size in pool %(pool)s',
                  {'name': name, 'size': size, 'pool': self.pool_id})
        self.tatlin_api.create_volume(volume.name_id,
                                      name,
                                      size,
                                      self.pool_id,
                                      lbaFormat=self._lba_format)

        self.wait_volume_ready(volume)

        self._update_qos(volume)

    def wait_volume_ready(self, volume):
        for counter in range(self._wait_retry_count):
            if self.tatlin_api.is_volume_ready(volume.name_id):
                return
            LOG.warning('Volume %s is not ready', volume.name_id)
            time.sleep(self._wait_interval)
        message = _('Volume %s still not ready') % volume.name_id
        LOG.error(message)
        raise exception.VolumeBackendAPIException(message=message)

    def wait_volume_online(self, volume):
        for counter in range(self._wait_retry_count):
            if self.tatlin_api.get_volume_status(volume.name_id) == 'online':
                return
            LOG.warning('Volume %s still not online', volume.name_id)
            time.sleep(self._wait_interval)
        message = _('Volume %s unable to become online' % volume.name_id)
        raise exception.VolumeBackendAPIException(message=message)

    @volume_utils.trace
    def delete_volume(self, volume):
        """Entry point for delete volume"""
        LOG.debug('Delete volume started for %s', volume.name_id)
        if not self.tatlin_api.is_volume_exists(volume.name_id):
            LOG.debug('Volume %s does not exist', volume.name_id)
            return
        try:
            self.tatlin_api.delete_volume(volume.name_id)
        except TatlinAPIException as e:
            message = _('Unable to delete volume %s due to %s' %
                        (volume.name_id, e))
            raise exception.VolumeBackendAPIException(message=message)

        for counter in range(self._wait_retry_count):
            if not self.tatlin_api.is_volume_exists(volume.name_id):
                LOG.debug('Delete volume finished for %s', volume.name_id)
                return
            LOG.debug('Volume %s still exists, waiting for delete...',
                      volume.name_id)
            time.sleep(self._wait_interval)

        if self.tatlin_api.is_volume_exists(volume.name_id):
            message = _('Unable to delete volume %s' % volume.name_id)
            LOG.error(message)
            raise exception.VolumeBackendAPIException(message=message)

    @volume_utils.trace
    def extend_volume(self, volume, new_size):
        size = new_size * units.Gi
        LOG.debug('Extending volume %s to %s', volume.name_id, size)
        self.tatlin_api.extend_volume(volume.name_id, size)
        self.wait_volume_ready(volume)
        self._update_qos(volume)

    @volume_utils.trace
    def create_cloned_volume(self, volume, src_vol):
        """Entry point for clone existing volume"""
        LOG.debug('Create cloned volume %(target)s from %(source)s started',
                  {'target': volume.name_id, 'source': src_vol.name_id})
        self.create_volume(volume)
        self._clone_volume_data(volume, src_vol)
        LOG.debug('Create cloned volume %(target)s from %(source)s finished',
                  {'target': volume.name_id, 'source': src_vol.name_id})

    def _clone_volume_data(self, volume, src_vol):
        props = brick_get_connector_properties(
            self._use_multipath,
            self._enforce_multipath)

        LOG.debug('Volume %s Connection properties %s',
                  volume.name_id, props)
        dest_attach_info = None
        src_attach_info = None

        size_in_mb = int(src_vol['size']) * units.Ki

        try:
            src_attach_info, volume_src = self._attach_volume(
                self.ctx, src_vol, props)
            LOG.debug('Source attach info: %s volume: %s',
                      src_attach_info, volume_src)

        except Exception as e:
            LOG.error('Unable to attach src volume due to %s', e)
            raise

        try:
            dest_attach_info, volume_dest = self._attach_volume(
                self.ctx, volume, props)
            LOG.debug('Dst attach info: %s volume: %s',
                      dest_attach_info, volume_dest)

        except Exception as e:
            LOG.error('Unable to attach dst volume due to %s', e)
            self._detach_volume(self.ctx, src_attach_info, src_vol, props)
            raise

        try:
            LOG.debug('Begin copy to %s from %s',
                      volume.name_id, src_vol.name_id)
            volume_utils.copy_volume(src_attach_info['device']['path'],
                                     dest_attach_info['device']['path'],
                                     size_in_mb,
                                     self.configuration.volume_dd_blocksize,
                                     sparse=False)
            LOG.debug('End copy to %s from %s',
                      volume.name_id, src_vol.name_id)
        except Exception as e:
            LOG.error('Unable to clone volume source: %s dst: %s due to %s',
                      src_vol.name_id, volume.name_id, e)
            raise
        finally:
            try:
                self._detach_volume(self.ctx, src_attach_info, src_vol, props)
            finally:
                self._detach_volume(self.ctx, dest_attach_info, volume, props)

    @volume_utils.trace
    def _attach_volume(self, context, volume, properties, remote=False):
        @utils.synchronized('tatlin-volume-attachments-%s' % volume.name_id)
        def _do_attach_volume():
            LOG.debug('Start Tatlin attach volume %s properties %s',
                      volume.name_id, properties)
            return super(driver.VolumeDriver, self)._attach_volume(
                context, volume, properties, remote=remote)
        return _do_attach_volume()

    @volume_utils.trace
    def _detach_volume(self, context, attach_info, volume, properties,
                       force=False, remote=False, ignore_errors=False):
        @utils.synchronized('tatlin-volume-attachments-%s' % volume.name_id)
        def _do_detach_volume():
            LOG.debug('Start Tatlin detach for %s', volume.name_id)
            connection_count = self._connections.get(volume.name_id)
            if connection_count > 1:
                LOG.debug('There are still other connections to volume %s,'
                          ' not detaching', volume.name_id)
                self._connections.decrement(volume.name_id)
                return
            # decrement of connections will happen in terminate_connection()
            super(driver.VolumeDriver, self).\
                _detach_volume(context, attach_info, volume, properties,
                               force=force,
                               remote=remote,
                               ignore_errors=ignore_errors)
        _do_detach_volume()

    @volume_utils.trace
    def terminate_connection(self, volume, connector, **kwargs):
        @utils.synchronized("tatlin-volume-connections-%s" % volume.name_id)
        def _terminate_connection():
            LOG.debug('Terminate connection for %s with connector  %s',
                      volume.name_id, connector)
            if not connector:
                return
            if self._is_cinder_host_connection(connector):
                connections = self._connections.decrement(volume.name_id)
                if connections > 0:
                    LOG.debug('Not terminating connection: '
                              'volume %s, existing connections: %s',
                              volume.name_id, connections)
                    return
            hostname = connector['host']
            if self._is_nova_multiattached(volume, hostname):
                LOG.debug('Volume %s is attached on host %s to multiple VMs.'
                          ' Not terminating connection', volume.name_id,
                          hostname)
                return

            host = self.find_current_host(connector['initiator'])
            LOG.debug('Terminate connection volume %s removing from host %s',
                      volume.name_id, host)
            self.remove_volume_from_host(volume, host)
        _terminate_connection()

    def _is_cinder_host_connection(self, connector):
        # Check if attachment happens on this Cinder host
        properties = brick_get_connector_properties()
        return properties['initiator'] == connector['initiator']

    def _is_nova_multiattached(self, volume, hostname):
        # Check if connection to the volume happens to multiple VMs
        # on the same Nova Compute host
        if not volume.volume_attachment:
            return False
        attachments = [a for a in volume.volume_attachment
                       if a.attach_status ==
                       objects.fields.VolumeAttachStatus.ATTACHED and
                       a.attached_host == hostname]
        return len(attachments) > 1

    def _create_temp_volume_for_snapshot(self, snapshot):
        return self._create_temp_volume(
            self.ctx,
            snapshot.volume,
            {
                'name_id': snapshot.id,
                'display_name': 'snap-vol-%s' % snapshot.id,
                'metadata': {'snapshot_volume': 'yes'},
            })

    @volume_utils.trace
    def create_snapshot(self, snapshot):
        LOG.debug('Create snapshot for volume %s, snap id %s',
                  snapshot.volume.name_id,
                  snapshot.id)
        temp_volume = self._create_temp_volume_for_snapshot(snapshot)
        try:
            self.create_cloned_volume(temp_volume, snapshot.volume)
        finally:
            temp_volume.destroy()

    @volume_utils.trace
    def create_volume_from_snapshot(self, volume, snapshot):
        LOG.debug('Create volume from snapshot %s', snapshot.id)
        temp_volume = self._create_temp_volume_for_snapshot(snapshot)
        try:
            self.create_volume(volume)
            self._clone_volume_data(volume, temp_volume)
        finally:
            temp_volume.destroy()

    @volume_utils.trace
    def delete_snapshot(self, snapshot):
        LOG.debug('Delete snapshot %s', snapshot.id)
        temp_volume = self._create_temp_volume_for_snapshot(snapshot)
        try:
            self.delete_volume(temp_volume)
        finally:
            temp_volume.destroy()

    @volume_utils.trace
    def get_volume_stats(self, refresh=False):
        if not self._stats or refresh:
            self._update_volume_stats()

        return self._stats

    def _update_qos(self, volume):
        type_id = volume.volume_type_id
        LOG.debug('VOL_TYPE %s', type_id)
        if type_id:
            ctx = cinder_context.get_admin_context()
            volume_type = volume_types.get_volume_type(ctx, type_id)
            qos_specs_id = volume_type.get('qos_specs_id')
            LOG.debug('VOL_TYPE %s QOS_SPEC %s', volume_type, qos_specs_id)
            specs = {}
            if qos_specs_id is not None:
                sp = qos_specs.get_qos_specs(ctx, qos_specs_id)
                if sp.get('consumer') != 'front-end':
                    specs = qos_specs.get_qos_specs(ctx, qos_specs_id)['specs']
            LOG.debug('QoS spec: %s', specs)
            param_specs = volume_type.get('extra_specs')
            LOG.debug('Param spec is: %s', param_specs)

            iops = specs["total_iops_sec_max"] \
                if 'total_iops_sec_max' in specs \
                else param_specs["YADRO:total_iops_sec_max"] \
                if 'YADRO:total_iops_sec_max' in param_specs else '0'

            bandwidth = specs["total_bytes_sec_max"] \
                if 'total_bytes_sec_max' in specs \
                else param_specs["YADRO:total_bytes_sec_max"] \
                if 'YADRO:total_bytes_sec_max' in param_specs else '0'

            LOG.debug('QOS spec IOPS: %s BANDWIDTH %s', iops, bandwidth)

            self.tatlin_api.update_qos(
                volume.name_id, int(iops), int(bandwidth))

    @volume_utils.trace
    def _update_volume_stats(self):
        """Retrieve pool info"""
        LOG.debug('Update volume stats for pool: %s', self.pool_name)
        if not self.pool_id:
            LOG.error('Could not retrieve pool id for %s', self.pool_name)
            return
        try:
            pool_stat = self.tatlin_api.get_pool_detail(self.pool_id)
        except TatlinAPIException as exp:
            message = (_('TatlinVolumeDriver get volume stats '
                       'failed %s due to %s') %
                       (self.pool_name, exp.message))
            LOG.error(message)
            return

        try:
            sys_stat = self.tatlin_api.get_sys_statistic()
        except TatlinAPIException as exp:
            message = (_('TatlinVolumeDriver get system stats detail '
                       'failed %s due to %s') %
                       (self.pool_name, exp.message))
            LOG.error(message)
            return

        if sys_stat['iops_bandwidth'] is not None and \
                len(sys_stat['iops_bandwidth']) > 0:
            self._stats['read_iops'] = \
                sys_stat['iops_bandwidth'][0]['value']['read_iops']
            self._stats['write_iops'] = \
                sys_stat['iops_bandwidth'][0]['value']['write_iops']
            self._stats['total_iops'] = \
                sys_stat['iops_bandwidth'][0]['value']['total_iops']
            self._stats['read_bytes_ps'] = \
                sys_stat['iops_bandwidth'][0]['value']['read_bytes_ps']
            self._stats['write_bytes_ps'] = \
                sys_stat['iops_bandwidth'][0]['value']['write_bytes_ps']
            self._stats['total_bytes_ps'] = \
                sys_stat['iops_bandwidth'][0]['value']['total_bytes_ps']

        self._stats["volume_backend_name"] = self.backend_name
        self._stats["vendor_name"] = 'YADRO'
        self._stats["driver_version"] = self.VERSION
        self._stats["storage_protocol"] = self.DRIVER_VOLUME_TYPE
        self._stats["thin_provisioning_support"] = pool_stat['thinProvision']
        self._stats["consistencygroup_support"] = False
        self._stats["consistent_group_snapshot_enabled"] = False
        self._stats["QoS_support"] = True
        self._stats["multiattach"] = True
        self._stats['total_capacity_gb'] = \
            (int(pool_stat['capacity']) - int(pool_stat['failed'])) / units.Gi
        self._stats['tatlin_pool'] = self.pool_name
        self._stats['tatlin_ip'] = self._ip
        pool_res_count, cluster_res_count = \
            self.tatlin_api.get_resource_count(self.pool_id)
        self._stats['overall_resource_count'] = cluster_res_count
        self._stats['pool_resource_count'] = pool_res_count
        if pool_stat['thinProvision']:
            self._stats['provisioned_capacity_gb'] = \
                (int(pool_stat['capacity']) -
                 int(pool_stat['failed'])) / units.Gi
            self._stats['free_capacity_gb'] = \
                self._stats['provisioned_capacity_gb']
        else:
            self._stats['provisioned_capacity_gb'] = \
                (int(pool_stat['available']) -
                 int(pool_stat['failed'])) / units.Gi
            self._stats['free_capacity_gb'] = \
                self._stats['provisioned_capacity_gb']

        self._stats['utilization'] = \
            (float(self._stats['total_capacity_gb']) -
             float(self._stats['free_capacity_gb'])) / \
            float(self._stats['total_capacity_gb']) * 100

        LOG.debug(
            'Total capacity: %s Free capacity: %s '
            'Provisioned capacity: %s '
            'Thin provisioning: %s '
            'Resource count: %s '
            'Pool resource count %s '
            'Utilization %s',
            self._stats['total_capacity_gb'],
            self._stats['free_capacity_gb'],
            self._stats['provisioned_capacity_gb'],
            pool_stat['thinProvision'], self._stats['overall_resource_count'],
            self._stats['pool_resource_count'],
            self._stats['utilization'])

    def _init_vendor_properties(self):
        LOG.debug('Initializing YADRO vendor properties')
        properties = {}
        self._set_property(
            properties,
            "YADRO:total_bytes_sec_max",
            "YADRO QoS Max bytes Write",
            _("Max write iops setting for volume qos, "
              "use 0 for unlimited"),
            "integer",
            minimum=0,
            default=0)
        self._set_property(
            properties,
            "YADRO:total_iops_sec_max",
            "YADRO QoS Max IOPS Write",
            _("Max write iops setting for volume qos, "
              "use 0 for unlimited"),
            "integer",
            minimum=0,
            default=0)
        LOG.debug('YADRO vendor properties: %s', properties)
        return properties, 'YADRO'

    def migrate_volume(self, context, volume, host):
        """Migrate volume

        Method checks if target volume will be on the same Tatlin/Pool
        If not, re-type should be executed.
        """
        if 'tatlin_pool' not in host['capabilities']:
            return False, None
        self._update_qos(volume)
        LOG.debug('Migrating volume from pool %s ip %s to pool %s ip %s',
                  self.pool_name, self._ip,
                  host['capabilities']['tatlin_pool'],
                  host['capabilities']['tatlin_ip'])
        if host['capabilities']['tatlin_ip'] == self._ip and \
                host['capabilities']['tatlin_pool'] == self.pool_name:
            return True, None

        return False, None

    def manage_existing(self, volume, external_ref):
        """Entry point to manage existing resource"""
        source_name = external_ref.get('source-name', None)
        if source_name is None:
            raise exception.ManageExistingInvalidReference(
                _('source_name should be provided'))
        try:
            result = self.tatlin_api.get_volume_info(source_name)
        except Exception:
            raise exception.ManageExistingInvalidReference(
                _('Unable to get resource with %s name' % source_name))

        existing_vol = result[0]
        existing_vol['name'] = volume.name_id
        volume.name_id = existing_vol['id']

        pool_id = existing_vol['poolId']

        if pool_id != self.pool_id:
            raise exception.ManageExistingInvalidReference(
                _('Existing volume should be in %s pool' % self.pool_name))

        self._update_qos(volume)

    def manage_existing_get_size(self, volume, external_ref):
        source_name = external_ref.get('source-name', None)
        if source_name is None:
            raise exception.ManageExistingInvalidReference(
                _('source_name should be provided'))

        try:
            result = self.tatlin_api.get_volume_info(source_name)
        except TatlinAPIException:
            raise exception.ManageExistingInvalidReference(
                _('Unable to get resource with %s name' % source_name))

        size = int(result[0]['size']) / units.G
        return size

    def add_volume_to_host(self, volume, host_id):
        self.tatlin_api.add_vol_to_host(volume.name_id, host_id)
        self._update_qos(volume)

    def remove_volume_from_host(self, volume, host_id):
        self.tatlin_api.remove_vol_from_host(volume.name_id, host_id)

    def _is_port_assigned(self, volume_id, port):
        LOG.debug('VOLUME %s: Checking port %s ', volume_id, port)
        cur_ports = self.tatlin_api.get_resource_ports_array(volume_id)
        res = port in cur_ports
        LOG.debug('VOLUME %s: port %s assigned %s',
                  volume_id, port, str(res))
        return res

    def _get_ports_portals(self):
        return {}

    def _find_mapped_lun(self, volume_id, iqn):
        host_id = self.find_current_host(iqn)
        result = self.tatlin_api.get_resource_mapping()
        for r in result:
            if 'host_id' in r:
                if r['resource_id'] == volume_id and r['host_id'] == host_id:
                    LOG.debug('Current mapped lun record %s volume_id: %s '
                              'host_id: is %s', r, volume_id, host_id)
                    return r['mapped_lun_id']

        mess = (_('Unable to get mapped lun for volume %s on host %s') %
                 (volume_id, host_id))
        LOG.error(mess)

        raise exception.VolumeBackendAPIException(message=mess)

    @staticmethod
    def get_driver_options():
        return tatlin_opts

    @volume_utils.trace
    def ensure_export(self, context, volume):
        LOG.debug('Tatlin ensure export')
        ports = self._get_ports_portals()
        self.tatlin_api.export_volume(volume.name_id, ports)

    @volume_utils.trace
    def create_export(self, context, volume, connector):
        LOG.debug('Create export for %s started', volume.name_id)
        self.ensure_export(context, volume)
        LOG.debug('Create export for %s finished', volume.name_id)

    def remove_export(self, context, volume):
        return

    def _get_tatlin_client(self):
        return InitTatlinClient(
            self._ip, self._port, self._user,
            self._password, verify=self.verify,
            api_retry_count=self.configuration.tat_api_retry_count,
            wait_interval=self._wait_interval,
            wait_retry_count=self._wait_retry_count)

    def find_current_host(self, wwn):
        return ''

    @property
    def pool_id(self):
        if not self._pool_id:
            try:
                self._pool_id = self.tatlin_api.get_pool_id_by_name(
                    self.pool_name)
            except exception.VolumeBackendAPIException:
                LOG.error('Unable to get current Tatlin pool')
        return self._pool_id

    @pool_id.setter
    def pool_id(self, value):
        self._pool_id = value

    @property
    def pool_name(self):
        return self._pool_name

    @pool_name.setter
    def pool_name(self, value):
        self._pool_name = value

    def get_default_filter_function(self):
        return self.DEFAULT_FILTER_FUNCTION

    def get_default_goodness_function(self):
        return self.DEFAULT_GOODNESS_FUNCTION

    @volume_utils.trace
    def get_backup_device(self, context, backup):
        """Get a backup device from an existing volume.

        We currently return original device where possible
        due to absence of instant clones and snapshots
        """
        if backup.snapshot_id:
            return super().get_backup_device(context, backup)

        volume = objects.Volume.get_by_id(context, backup.volume_id)
        return (volume, False)

    def backup_use_temp_snapshot(self):
        return False

    def snapshot_revert_use_temp_snapshot(self):
        return False



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils

from cinder.tests.functional import api_samples_test_base as test_base


class VolumeTransferSampleJsonTest(test_base.VolumesSampleBase):
    sample_dir = "volume_transfer"

    def setUp(self):
        super(VolumeTransferSampleJsonTest, self).setUp()
        res = self._create_volume()
        res = jsonutils.loads(res.content)['volume']
        self._poll_volume_while(res['id'], ['creating'])
        self.subs = {
            "volume_id": res['id']
        }
        self.response = self._create_transfer(self.subs)

    def _create_transfer(self, subs=None):
        response = self._do_post('os-volume-transfer',
                                 'volume-transfer-create-request',
                                 subs)
        return response

    def test_transfer_create(self):

        self._verify_response('volume-transfer-create-response',
                              {}, self.response, 202)

    def test_transfer_accept(self):

        res = jsonutils.loads(self.response.content)['transfer']
        subs = {
            "auth_key": res['auth_key']
        }
        response = self._do_post(
            'os-volume-transfer/%s/accept' % res['id'],
            'volume-transfer-accept-request',
            subs)
        self._verify_response('volume-transfer-accept-response',
                              {}, response, 202)

    def test_transfers_list(self):

        response = self._do_get('os-volume-transfer')
        self._verify_response('volume-transfers-list-response',
                              {}, response, 200)

    def test_transfer_list_detail(self):

        res = jsonutils.loads(self.response.content)['transfer']
        response = self._do_get('os-volume-transfer/%s' % res['id'])
        self._verify_response('volume-transfer-show-response',
                              {}, response, 200)

    def test_transfers_list_detail(self):

        response = self._do_get('os-volume-transfer/detail')
        self._verify_response('volume-transfers-list-detailed-response',
                              {}, response, 200)



# Copyright 2016 Dell Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""
Manage/unmanage existing volume snapshots driver interface.
"""

from cinder.interface import base


class VolumeSnapshotManagementDriver(base.CinderInterface):
    """Interface for drivers that support managing existing snapshots."""

    def manage_existing_snapshot(self, snapshot, existing_ref):
        """Brings an existing backend storage object under Cinder management.

        existing_ref is passed straight through from the API request's
        manage_existing_ref value, and it is up to the driver how this should
        be interpreted.  It should be sufficient to identify a storage object
        that the driver should somehow associate with the newly-created cinder
        snapshot structure.

        There are two ways to do this:

        1. Rename the backend storage object so that it matches the
           snapshot['name'] which is how drivers traditionally map between a
           cinder snapshot and the associated backend storage object.

        2. Place some metadata on the snapshot, or somewhere in the backend,
           that allows other driver requests (e.g. delete) to locate the
           backend storage object when required.

        :param snapshot: The snapshot to manage.
        :param existing_ref: Dictionary with keys 'source-id', 'source-name'
                             with driver-specific values to identify a backend
                             storage object.
        :raises ManageExistingInvalidReference: If the existing_ref doesn't
                 make sense, or doesn't refer to an existing backend storage
                 object.
        """

    def manage_existing_snapshot_get_size(self, snapshot, existing_ref):
        """Return size of snapshot to be managed by manage_existing.

        When calculating the size, round up to the next GB.

        :param snapshot: The snapshot to manage.
        :param existing_ref: Dictionary with keys 'source-id', 'source-name'
                             with driver-specific values to identify a backend
                             storage object.
        :raises ManageExistingInvalidReference: If the existing_ref doesn't
                 make sense, or doesn't refer to an existing backend storage
                 object.
        """

    def unmanage_snapshot(self, snapshot):
        """Removes the specified snapshot from Cinder management.

        Does not delete the underlying backend storage object.

        For most drivers, this will not need to do anything. However, some
        drivers might use this call as an opportunity to clean up any
        Cinder-specific configuration that they have associated with the
        backend storage object.

        :param snapshot: The snapshot to unmanage.
        """



# Copyright (c) 2016 EMC Corporation, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from os import path
from unittest import mock

import six
import yaml

from cinder.volume.drivers.dell_emc.vnx import client
from cinder.volume.drivers.dell_emc.vnx import common


patch_sleep = mock.patch('time.sleep')


patch_vnxsystem = mock.patch('storops.VNXSystem')


patch_vnxstoragegroup = mock.patch('storops.vnx.resource.sg.VNXStorageGroup')


def load_yaml(file_name):
    yaml_file = '{}/{}'.format(path.dirname(
        path.abspath(__file__)), file_name)
    with open(yaml_file) as f:
        res = yaml.safe_load(f)
    return res


def patch_extra_specs(specs):
    return _build_patch_decorator(
        'cinder.volume.volume_types.get_volume_type_extra_specs',
        return_value=specs)


def patch_group_specs(specs):
    return _build_patch_decorator(
        'cinder.volume.group_types.get_group_type_specs',
        return_value=specs)


def patch_extra_specs_validate(return_value=None, side_effect=None):
    return _build_patch_decorator(
        'cinder.volume.drivers.dell_emc.vnx.common.ExtraSpecs.validate',
        return_value=return_value,
        side_effect=side_effect)


def _build_patch_decorator(module_str, return_value=None, side_effect=None):
    def _inner_mock(func):
        @six.wraps(func)
        def decorator(*args, **kwargs):
            with mock.patch(
                    module_str,
                    return_value=return_value,
                    side_effect=side_effect):
                return func(*args, **kwargs)
        return decorator
    return _inner_mock


def build_fake_mirror_view():
    primary_client = mock.create_autospec(spec=client.Client)
    secondary_client = mock.create_autospec(spec=client.Client)

    mirror_view = mock.create_autospec(spec=common.VNXMirrorView)
    mirror_view.primary_client = primary_client
    mirror_view.secondary_client = secondary_client
    return mirror_view


def get_replication_device():
    return {
        'backend_id': 'fake_serial',
        'san_ip': '192.168.1.12',
        'san_login': 'admin',
        'san_password': 'admin',
        'storage_vnx_authentication_type': 'global',
        'storage_vnx_security_file_dir': None,
        'pool_name': 'remote_pool',
    }



# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import six

from cinder import context
from cinder.tests.unit.consistencygroup import fake_cgsnapshot
from cinder.tests.unit.consistencygroup import fake_consistencygroup
from cinder.tests.unit import fake_constants
from cinder.tests.unit import fake_group
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit.volume.drivers.dell_emc.vnx import fake_exception as \
    lib_ex
from cinder.tests.unit.volume.drivers.dell_emc.vnx import fake_storops as \
    storops
from cinder.tests.unit.volume.drivers.dell_emc.vnx import utils
from cinder.volume.drivers.dell_emc.vnx import adapter
from cinder.volume.drivers.dell_emc.vnx import client
from cinder.volume.drivers.dell_emc.vnx import common
from cinder.volume.drivers.dell_emc.vnx import driver
from cinder.volume.drivers.dell_emc.vnx import utils as vnx_utils

SYMBOL_TYPE = '_type'
SYMBOL_PROPERTIES = '_properties'
SYMBOL_METHODS = '_methods'
SYMBOL_SIDE_EFFECT = '_side_effect'
SYMBOL_RAISE = '_raise'
SYMBOL_CONTEXT = '_context'
UUID = '_uuid'
SYMBOL_ENUM = '_enum'


def _is_driver_object(obj_body):
    return isinstance(obj_body, dict) and SYMBOL_PROPERTIES in obj_body


class DriverResourceMock(dict):
    fake_func_mapping = {}

    def __init__(self, yaml_file):
        yaml_dict = utils.load_yaml(yaml_file)
        if not isinstance(yaml_dict, dict):
            return
        for case_name, case_res in yaml_dict.items():
            if not isinstance(case_res, dict):
                continue
            self[case_name] = {}
            for obj_name, obj_body in case_res.items():
                self[case_name][obj_name] = self._parse_driver_object(obj_body)

    def _parse_driver_object(self, obj_body):
        if isinstance(obj_body, dict):
            obj_body = {k: self._parse_driver_object(v)
                        for k, v in obj_body.items()}
            if _is_driver_object(obj_body):
                return self._create_object(obj_body)
            else:
                return obj_body
        elif isinstance(obj_body, list):
            return map(self._parse_driver_object, obj_body)
        else:
            return obj_body

    def _create_object(self, obj_body):
        props = obj_body[SYMBOL_PROPERTIES]
        for prop_name, prop_value in props.items():
            if isinstance(prop_value, dict) and prop_value:
                # get the first key as the convert function
                func_name = list(prop_value.keys())[0]
                if func_name.startswith('_'):
                    func = getattr(self, func_name)
                    props[prop_name] = func(prop_value[func_name])

        if (SYMBOL_TYPE in obj_body and
                obj_body[SYMBOL_TYPE] in self.fake_func_mapping):
            return self.fake_func_mapping[obj_body[SYMBOL_TYPE]](**props)
        else:
            return props

    @staticmethod
    def _uuid(uuid_key):
        uuid_key = uuid_key.upper()
        return getattr(fake_constants, uuid_key)


def _fake_volume_wrapper(*args, **kwargs):
    expected_attrs_key = {'volume_attachment': 'volume_attachment',
                          'volume_metadata': 'metadata'}
    if 'group' in kwargs:
        expected_attrs_key['group'] = kwargs['group']

    return fake_volume.fake_volume_obj(
        context.get_admin_context(),
        expected_attrs=[
            v for (k, v) in expected_attrs_key.items() if k in kwargs],
        **kwargs)


def _fake_cg_wrapper(*args, **kwargs):
    return fake_consistencygroup.fake_consistencyobject_obj(
        'fake_context', **kwargs)


def _fake_snapshot_wrapper(*args, **kwargs):
    return fake_snapshot.fake_snapshot_obj('fake_context',
                                           expected_attrs=(
                                               ['volume'] if 'volume' in kwargs
                                               else None),
                                           **kwargs)


def _fake_cg_snapshot_wrapper(*args, **kwargs):
    return fake_cgsnapshot.fake_cgsnapshot_obj(None, **kwargs)


def _fake_group_wrapper(*args, **kwargs):
    return fake_group.fake_group_obj(None, **kwargs)


class EnumBuilder(object):
    def __init__(self, enum_dict):
        enum_dict = enum_dict[SYMBOL_ENUM]
        for k, v in enum_dict.items():
            self.klazz = k
            self.value = v

    def __call__(self, *args, **kwargs):
        return getattr(storops, self.klazz).parse(self.value)


class CinderResourceMock(DriverResourceMock):
    # fake_func in the mapping should be like func(*args, **kwargs)
    fake_func_mapping = {'volume': _fake_volume_wrapper,
                         'cg': _fake_cg_wrapper,
                         'snapshot': _fake_snapshot_wrapper,
                         'cg_snapshot': _fake_cg_snapshot_wrapper,
                         'group': _fake_group_wrapper}

    def __init__(self, yaml_file):
        super(CinderResourceMock, self).__init__(yaml_file)

    @staticmethod
    def _build_provider_location(props):
        return vnx_utils.build_provider_location(
            props.get('system'), props.get('type'),
            six.text_type(props.get('id')),
            six.text_type(props.get('base_lun_name')),
            props.get('version'))


class ContextMock(object):
    """Mocks the return value of a context function."""

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_valu, exc_tb):
        pass


class MockBase(object):
    """Base object of all the Mocks.

    This mock convert the dict to object when the '_type' is
    included in the dict
    """

    def _is_mock_object(self, yaml_info):
        return (isinstance(yaml_info, dict) and
                (SYMBOL_PROPERTIES in yaml_info or
                 SYMBOL_METHODS in yaml_info))

    def _is_object_with_type(self, yaml_dict):
        return isinstance(yaml_dict, dict) and SYMBOL_TYPE in yaml_dict

    def _is_object_with_enum(self, yaml_dict):
        return isinstance(yaml_dict, dict) and SYMBOL_ENUM in yaml_dict

    def _build_mock_object(self, yaml_dict):
        if self._is_object_with_type(yaml_dict):
            return FakePort(yaml_dict)
        elif self._is_object_with_enum(yaml_dict):
            return EnumBuilder(yaml_dict)()
        elif self._is_mock_object(yaml_dict):
            return StorageObjectMock(yaml_dict)
        elif isinstance(yaml_dict, dict):
            return {k: self._build_mock_object(v)
                    for k, v in yaml_dict.items()}
        elif isinstance(yaml_dict, list):
            return [self._build_mock_object(each) for each in yaml_dict]
        else:
            return yaml_dict


class StorageObjectMock(object):
    PROPS = 'props'

    def __init__(self, yaml_dict):
        self.__dict__[StorageObjectMock.PROPS] = {}
        props = yaml_dict.get(SYMBOL_PROPERTIES, None)
        if props:
            for k, v in props.items():
                setattr(self, k, StoragePropertyMock(k, v)())

        methods = yaml_dict.get(SYMBOL_METHODS, None)
        if methods:
            for k, v in methods.items():
                setattr(self, k, StorageMethodMock(k, v))

    def __setattr__(self, key, value):
        self.__dict__[StorageObjectMock.PROPS][key] = value

    def __getattr__(self, item):
        try:
            super(StorageObjectMock, self).__getattr__(item)
        except AttributeError:
            return self.__dict__[StorageObjectMock.PROPS][item]
        except KeyError:
            raise KeyError('%(item)s not exist in mock object.'
                           ) % {'item': item}


class FakePort(StorageObjectMock):

    def __eq__(self, other):
        o_sp = other.sp
        o_port_id = other.port_id
        o_vport_id = other.vport_id

        ret = True
        ret &= self.sp == o_sp
        ret &= self.port_id == o_port_id
        ret &= self.vport_id == o_vport_id

        return ret

    def __hash__(self):
        return hash((self.sp, self.port_id, self.vport_id))


class StoragePropertyMock(mock.PropertyMock, MockBase):
    def __init__(self, name, property_body):
        return_value = property_body
        side_effect = None

        # only support return_value and side_effect for property
        if (isinstance(property_body, dict) and
                SYMBOL_SIDE_EFFECT in property_body):
            side_effect = self._build_mock_object(
                property_body[SYMBOL_SIDE_EFFECT])
            return_value = None

        if side_effect is not None:
            super(StoragePropertyMock, self).__init__(
                name=name,
                side_effect=side_effect)
        else:
            return_value = self._build_mock_object(return_value)

            super(StoragePropertyMock, self).__init__(
                name=name,
                return_value=return_value)


class StorageMethodMock(mock.Mock, MockBase):
    def __init__(self, name, method_body):
        return_value = method_body
        exception = None
        side_effect = None

        # support return_value, side_effect and exception for method
        if isinstance(method_body, dict):
            if (SYMBOL_SIDE_EFFECT in method_body or
                    SYMBOL_RAISE in method_body):
                exception = method_body.get(SYMBOL_RAISE, None)
                side_effect = method_body.get(SYMBOL_SIDE_EFFECT, None)
                return_value = None

        if exception is not None:
            ex = None
            if isinstance(exception, dict) and exception:
                ex_name = list(exception.keys())[0]
                ex_tmp = [getattr(ex_module, ex_name, None)
                          for ex_module in [lib_ex, common]]
                try:
                    ex = [each for each in ex_tmp if each is not None][0]
                    super(StorageMethodMock, self).__init__(
                        name=name,
                        side_effect=ex(exception[ex_name]))
                except IndexError:
                    raise KeyError('Exception %(ex_name)s not found.'
                                   % {'ex_name': ex_name})
            else:
                raise KeyError('Invalid Exception body, should be a dict.')
        elif side_effect is not None:
            super(StorageMethodMock, self).__init__(
                name=name,
                side_effect=self._build_mock_object(side_effect))
        elif return_value is not None:
            super(StorageMethodMock, self).__init__(
                name=name,
                return_value=(ContextMock() if return_value == SYMBOL_CONTEXT
                              else self._build_mock_object(return_value)))
        else:
            super(StorageMethodMock, self).__init__(
                name=name, return_value=None)


class StorageResourceMock(dict, MockBase):
    def __init__(self, yaml_file):
        yaml_dict = utils.load_yaml(yaml_file)
        if not isinstance(yaml_dict, dict):
            return
        for section, sec_body in yaml_dict.items():
            if isinstance(sec_body, dict):
                self[section] = {obj_name: self._build_mock_object(obj_body)
                                 for obj_name, obj_body
                                 in sec_body.items()}
            else:
                self[section] = {}


cinder_res = CinderResourceMock('mocked_cinder.yaml')
DRIVER_RES_MAPPING = {
    'TestResMock': cinder_res,
    'TestCommonAdapter': cinder_res,
    'TestReplicationAdapter': cinder_res,
    'TestISCSIAdapter': cinder_res,
    'TestFCAdapter': cinder_res,
    'TestUtils': cinder_res,
    'TestClient': cinder_res
}


def mock_driver_input(func):
    @six.wraps(func)
    def decorated(cls, *args, **kwargs):
        return func(cls,
                    DRIVER_RES_MAPPING[cls.__class__.__name__][func.__name__],
                    *args, **kwargs)
    return decorated


vnx_res = StorageResourceMock('mocked_vnx.yaml')
STORAGE_RES_MAPPING = {
    'TestResMock': StorageResourceMock('test_res_mock.yaml'),
    'TestCondition': vnx_res,
    'TestClient': vnx_res,
    'TestCommonAdapter': vnx_res,
    'TestReplicationAdapter': vnx_res,
    'TestISCSIAdapter': vnx_res,
    'TestFCAdapter': vnx_res,
    'TestTaskflow': vnx_res,
    'TestExtraSpecs': vnx_res,
}
DEFAULT_STORAGE_RES = 'vnx'


def _build_client():
    return client.Client(ip='192.168.1.2',
                         username='sysadmin',
                         password='sysadmin',
                         scope='global',
                         naviseccli=None,
                         sec_file=None,
                         queue_path='vnx-cinder')


def patch_client(func):
    @six.wraps(func)
    def decorated(cls, *args, **kwargs):
        storage_res = (
            STORAGE_RES_MAPPING[cls.__class__.__name__][func.__name__])
        with utils.patch_vnxsystem as patched_vnx:
            if DEFAULT_STORAGE_RES in storage_res:
                patched_vnx.return_value = storage_res[DEFAULT_STORAGE_RES]
            client = _build_client()
        return func(cls, client, storage_res, *args, **kwargs)
    return decorated


PROTOCOL_COMMON = 'Common'
PROTOCOL_MAPPING = {
    PROTOCOL_COMMON: adapter.CommonAdapter,
    common.PROTOCOL_ISCSI: adapter.ISCSIAdapter,
    common.PROTOCOL_FC: adapter.FCAdapter
}


def patch_adapter_init(protocol):
    def inner_patch_adapter(func):
        @six.wraps(func)
        def decorated(cls, *args, **kwargs):
            storage_res = (
                STORAGE_RES_MAPPING[cls.__class__.__name__][func.__name__])
            with utils.patch_vnxsystem as patched_vnx:
                if DEFAULT_STORAGE_RES in storage_res:
                    patched_vnx.return_value = storage_res[DEFAULT_STORAGE_RES]
                adapter = PROTOCOL_MAPPING[protocol](cls.configuration)
            return func(cls, adapter, storage_res, *args, **kwargs)
        return decorated
    return inner_patch_adapter


def _patch_adapter_prop(adapter, client):
    try:
        adapter.serial_number = client.get_serial()
    except KeyError:
        adapter.serial_number = 'faked_serial_number'
    adapter.VERSION = driver.VNXDriver.VERSION


def patch_adapter(protocol):
    def inner_patch_adapter(func):
        @six.wraps(func)
        def decorated(cls, *args, **kwargs):
            storage_res = (
                STORAGE_RES_MAPPING[cls.__class__.__name__][func.__name__])
            with utils.patch_vnxsystem:
                client = _build_client()
                adapter = PROTOCOL_MAPPING[protocol](cls.configuration, None)
            if DEFAULT_STORAGE_RES in storage_res:
                client.vnx = storage_res[DEFAULT_STORAGE_RES]
            adapter.client = client
            _patch_adapter_prop(adapter, client)
            return func(cls, adapter, storage_res, *args, **kwargs)
        return decorated
    return inner_patch_adapter


patch_common_adapter = patch_adapter(PROTOCOL_COMMON)
patch_iscsi_adapter = patch_adapter(common.PROTOCOL_ISCSI)
patch_fc_adapter = patch_adapter(common.PROTOCOL_FC)


def mock_storage_resources(func):
    @six.wraps(func)
    def decorated(cls, *args, **kwargs):
        storage_res = (
            STORAGE_RES_MAPPING[cls.__class__.__name__][func.__name__])
        return func(cls, storage_res, *args, **kwargs)
    return decorated



#  Copyright (C) 2021-2022 YADRO.
#  All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

HOST_GROUPS = 'personalities/v1/config/groups'
HOSTS = 'personalities/v1/config/hosts'
RESOURCE = 'personalities/v1/personalities/block/%s'
RESOURCE_DETAIL = 'personalities/v1/personalities?id=%s'
RESOURCE_HEALTH = 'health/v1/personalities?id=%s'
RESOURCE_MAPPING = 'personalities/v1/config/resource_mapping'
VOLUME_TO_HOST = 'personalities/v1/personalities/block/%s/hosts/%s'
ALL_RESOURCES = 'personalities/v1/personalities'
POOLS = 'health/v1/pools'
STATISTICS = 'health/v1/statistics/current'
IP_PORTS = 'osmgr/v1/ports/%s'
RESOURCE_COUNT = 'personalities/v1/personalities/block/countPerPool'
TATLIN_VERSION = 'upmgr/v1/version'



# Copyright (c) 2017 VMware, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
VMware VStorageObject driver

Volume driver based on VMware VStorageObject aka First Class Disk (FCD). This
driver requires a minimum vCenter version of 6.5.
"""

from oslo_log import log as logging
from oslo_utils import units
from oslo_utils import versionutils
from oslo_vmware import image_transfer
from oslo_vmware.objects import datastore
from oslo_vmware import vim_util

from cinder.common import constants
from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder.volume.drivers.vmware import datastore as hub
from cinder.volume.drivers.vmware import vmdk
from cinder.volume.drivers.vmware import volumeops as vops
from cinder.volume import volume_utils


LOG = logging.getLogger(__name__)


@interface.volumedriver
class VMwareVStorageObjectDriver(vmdk.VMwareVcVmdkDriver):
    """Volume driver based on VMware VStorageObject"""

    # 1.0 - initial version based on vSphere 6.5 vStorageObject APIs
    # 1.1 - support for vStorageObject snapshot APIs
    # 1.2 - support for SPBM storage policies
    # 1.3 - support for retype
    VERSION = '1.3.0'

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = "VMware_CI"

    # minimum supported vCenter version
    MIN_SUPPORTED_VC_VERSION = '6.5'

    STORAGE_TYPE = constants.VSTORAGE

    def do_setup(self, context):
        """Any initialization the volume driver needs to do while starting.

        :param context: The admin context.
        """
        super(VMwareVStorageObjectDriver, self).do_setup(context)
        self.volumeops.set_vmx_version('vmx-13')
        vc_67_compatible = versionutils.is_compatible(
            '6.7.0', self._vc_version, same_major=False)
        self._use_fcd_snapshot = vc_67_compatible
        self._storage_policy_enabled = vc_67_compatible

    def get_volume_stats(self, refresh=False):
        """Collects volume backend stats.

        :param refresh: Whether to discard any cached values and force a full
                        refresh of stats.
        :returns: dict of appropriate values.
        """
        stats = super(VMwareVStorageObjectDriver, self).get_volume_stats(
            refresh=refresh)
        stats['storage_protocol'] = self.STORAGE_TYPE
        return stats

    def _select_ds_fcd(self, volume):
        req = {}
        req[hub.DatastoreSelector.SIZE_BYTES] = volume.size * units.Gi

        if self._storage_policy_enabled:
            req[hub.DatastoreSelector.PROFILE_NAME] = (
                self._get_storage_profile(volume))
        (_host_ref, _resource_pool, summary) = self._select_datastore(req)
        return summary.datastore

    def _get_temp_image_folder(self, size_bytes, preallocated=False):
        req = {}
        req[hub.DatastoreSelector.SIZE_BYTES] = size_bytes

        if preallocated:
            req[hub.DatastoreSelector.HARD_AFFINITY_DS_TYPE] = (
                hub.DatastoreType.get_all_types() -
                {hub.DatastoreType.VSAN, hub.DatastoreType.VVOL})

        (host_ref, _resource_pool, summary) = self._select_datastore(req)

        folder_path = vmdk.TMP_IMAGES_DATASTORE_FOLDER_PATH
        dc_ref = self.volumeops.get_dc(host_ref)
        self.volumeops.create_datastore_folder(
            summary.name, folder_path, dc_ref)

        return (dc_ref, summary, folder_path)

    def _get_disk_type(self, volume):
        extra_spec_disk_type = super(
            VMwareVStorageObjectDriver, self)._get_disk_type(volume)
        return vops.VirtualDiskType.get_virtual_disk_type(extra_spec_disk_type)

    def _get_storage_profile_id(self, volume):
        if self._storage_policy_enabled:
            return super(
                VMwareVStorageObjectDriver, self)._get_storage_profile_id(
                    volume)

    def create_volume(self, volume):
        """Create a new volume on the backend.

        :param volume: Volume object containing specifics to create.
        :returns: (Optional) dict of database updates for the new volume.
        """
        disk_type = self._get_disk_type(volume)
        ds_ref = self._select_ds_fcd(volume)
        profile_id = self._get_storage_profile_id(volume)
        fcd_loc = self.volumeops.create_fcd(
            volume.name, volume.size * units.Ki, ds_ref, disk_type,
            profile_id=profile_id)
        return {'provider_location': fcd_loc.provider_location()}

    def _delete_fcd(self, provider_loc):
        fcd_loc = vops.FcdLocation.from_provider_location(provider_loc)
        self.volumeops.delete_fcd(fcd_loc)

    def delete_volume(self, volume):
        """Delete a volume from the backend.

        :param volume: The volume to delete.
        """
        if not volume.provider_location:
            LOG.warning("FCD provider location is empty for volume %s",
                        volume.id)
        else:
            self._delete_fcd(volume.provider_location)

    def initialize_connection(self, volume, connector, initiator_data=None):
        """Allow connection to connector and return connection info.

        :param volume: The volume to be attached.
        :param connector: Dictionary containing information about what is being
                          connected to.
        :param initiator_data: (Optional) A dictionary of driver_initiator_data
                               objects with key-value pairs that have been
                               saved for this initiator by a driver in previous
                               initialize_connection calls.
        :returns: A dictionary of connection information.
        """
        fcd_loc = vops.FcdLocation.from_provider_location(
            volume.provider_location)
        connection_info = {'driver_volume_type': self.STORAGE_TYPE}
        connection_info['data'] = {
            'id': fcd_loc.fcd_id,
            'ds_ref_val': fcd_loc.ds_ref_val,
            'adapter_type': self._get_adapter_type(volume)
        }
        LOG.debug("Connection info for volume %(name)s: %(connection_info)s.",
                  {'name': volume.name, 'connection_info': connection_info})
        return connection_info

    def _validate_container_format(self, container_format, image_id):
        if container_format and container_format != 'bare':
            msg = _("Container format: %s is unsupported, only 'bare' "
                    "is supported.") % container_format
            LOG.error(msg)
            raise exception.ImageUnacceptable(image_id=image_id, reason=msg)

    def copy_image_to_volume(self, context, volume, image_service, image_id):
        """Fetch the image from image_service and write it to the volume.

        :param context: Security/policy info for the request.
        :param volume: The volume to create.
        :param image_service: The image service to use.
        :param image_id: The image identifier.
        :returns: Model updates.
        """
        metadata = image_service.show(context, image_id)
        self._validate_disk_format(metadata['disk_format'])
        self._validate_container_format(
            metadata.get('container_format'), image_id)

        properties = metadata['properties'] or {}
        disk_type = properties.get('vmware_disktype',
                                   vmdk.ImageDiskType.PREALLOCATED)
        vmdk.ImageDiskType.validate(disk_type)

        size_bytes = metadata['size']
        dc_ref, summary, folder_path = self._get_temp_image_folder(
            volume.size * units.Gi)
        disk_name = volume.id
        if disk_type in [vmdk.ImageDiskType.SPARSE,
                         vmdk.ImageDiskType.STREAM_OPTIMIZED]:
            vmdk_path = self._create_virtual_disk_from_sparse_image(
                context, image_service, image_id, size_bytes, dc_ref,
                summary.name, folder_path, disk_name)
        else:
            vmdk_path = self._create_virtual_disk_from_preallocated_image(
                context, image_service, image_id, size_bytes, dc_ref,
                summary.name, folder_path, disk_name,
                vops.VirtualDiskAdapterType.LSI_LOGIC)

        ds_path = datastore.DatastorePath.parse(
            vmdk_path.get_descriptor_ds_file_path())
        dc_path = self.volumeops.get_inventory_path(dc_ref)

        vmdk_url = datastore.DatastoreURL(
            'https', self.configuration.vmware_host_ip, ds_path.rel_path,
            dc_path, ds_path.datastore)

        fcd_loc = self.volumeops.register_disk(
            str(vmdk_url), volume.name, summary.datastore)

        profile_id = self._get_storage_profile_id(volume)
        if profile_id:
            self.volumeops.update_fcd_policy(fcd_loc, profile_id)

        return {'provider_location': fcd_loc.provider_location()}

    def copy_volume_to_image(self, context, volume, image_service, image_meta):
        """Copy the volume to the specified image.

        :param context: Security/policy info for the request.
        :param volume: The volume to copy.
        :param image_service: The image service to use.
        :param image_meta: Information about the image.
        :returns: Model updates.
        """
        self._validate_disk_format(image_meta['disk_format'])

        fcd_loc = vops.FcdLocation.from_provider_location(
            volume.provider_location)
        hosts = self.volumeops.get_connected_hosts(fcd_loc.ds_ref())
        host = vim_util.get_moref(hosts[0], 'HostSystem')
        LOG.debug("Selected host: %(host)s for downloading fcd: %(fcd_loc)s.",
                  {'host': host, 'fcd_loc': fcd_loc})

        attached = False
        try:
            create_params = {vmdk.CREATE_PARAM_DISK_LESS: True}
            backing = self._create_backing(volume, host, create_params)
            self.volumeops.attach_fcd(backing, fcd_loc)
            attached = True

            vmdk_file_path = self.volumeops.get_vmdk_path(backing)
            conf = self.configuration

            # retrieve store information from extra-specs
            store_id = volume.volume_type.extra_specs.get(
                'image_service:store_id')

            # TODO (whoami-rajat): Remove store_id and base_image_ref
            #  parameters when oslo.vmware calls volume_utils wrapper of
            #  upload_volume instead of image_utils.upload_volume
            image_transfer.upload_image(
                context,
                conf.vmware_image_transfer_timeout_secs,
                image_service,
                image_meta['id'],
                volume.project_id,
                session=self.session,
                host=conf.vmware_host_ip,
                port=conf.vmware_host_port,
                vm=backing,
                vmdk_file_path=vmdk_file_path,
                vmdk_size=volume.size * units.Gi,
                image_name=image_meta['name'],
                store_id=store_id,
                base_image_ref=volume_utils.get_base_image_ref(volume))
        finally:
            if attached:
                self.volumeops.detach_fcd(backing, fcd_loc)
            backing = self.volumeops.get_backing_by_uuid(volume.id)
            if backing:
                self._delete_temp_backing(backing)

    def extend_volume(self, volume, new_size):
        """Extend the size of a volume.

        :param volume: The volume to extend.
        :param new_size: The new desired size of the volume.
        """
        fcd_loc = vops.FcdLocation.from_provider_location(
            volume.provider_location)
        self.volumeops.extend_fcd(fcd_loc, new_size * units.Ki)

    def _clone_fcd(self, provider_loc, name, dest_ds_ref,
                   disk_type=vops.VirtualDiskType.THIN,
                   profile_id=None):
        fcd_loc = vops.FcdLocation.from_provider_location(provider_loc)
        return self.volumeops.clone_fcd(
            name, fcd_loc, dest_ds_ref, disk_type, profile_id=profile_id)

    def create_snapshot(self, snapshot):
        """Creates a snapshot.

        :param snapshot: Information for the snapshot to be created.
        """
        if self._use_fcd_snapshot:
            fcd_loc = vops.FcdLocation.from_provider_location(
                snapshot.volume.provider_location)
            description = "snapshot-%s" % snapshot.id
            fcd_snap_loc = self.volumeops.create_fcd_snapshot(
                fcd_loc, description=description)
            return {'provider_location': fcd_snap_loc.provider_location()}

        ds_ref = self._select_ds_fcd(snapshot.volume)
        cloned_fcd_loc = self._clone_fcd(
            snapshot.volume.provider_location, snapshot.name, ds_ref)
        return {'provider_location': cloned_fcd_loc.provider_location()}

    def delete_snapshot(self, snapshot):
        """Deletes a snapshot.

        :param snapshot: The snapshot to delete.
        """
        if not snapshot.provider_location:
            LOG.debug("FCD snapshot location is empty.")
            return

        fcd_snap_loc = vops.FcdSnapshotLocation.from_provider_location(
            snapshot.provider_location)
        if fcd_snap_loc:
            self.volumeops.delete_fcd_snapshot(fcd_snap_loc)
        else:
            self._delete_fcd(snapshot.provider_location)

    def _extend_if_needed(self, fcd_loc, cur_size, new_size):
        if new_size > cur_size:
            self.volumeops.extend_fcd(fcd_loc, new_size * units.Ki)

    def _create_volume_from_fcd(self, provider_loc, cur_size, volume):
        ds_ref = self._select_ds_fcd(volume)
        disk_type = self._get_disk_type(volume)
        profile_id = self._get_storage_profile_id(volume)
        cloned_fcd_loc = self._clone_fcd(
            provider_loc, volume.name, ds_ref, disk_type=disk_type,
            profile_id=profile_id)
        self._extend_if_needed(cloned_fcd_loc, cur_size, volume.size)
        return {'provider_location': cloned_fcd_loc.provider_location()}

    def create_volume_from_snapshot(self, volume, snapshot):
        """Creates a volume from a snapshot.

        :param volume: The volume to be created.
        :param snapshot: The snapshot from which to create the volume.
        :returns: A dict of database updates for the new volume.
        """
        fcd_snap_loc = vops.FcdSnapshotLocation.from_provider_location(
            snapshot.provider_location)
        if fcd_snap_loc:
            profile_id = self._get_storage_profile_id(volume)
            fcd_loc = self.volumeops.create_fcd_from_snapshot(
                fcd_snap_loc, volume.name, profile_id=profile_id)
            self._extend_if_needed(fcd_loc, snapshot.volume_size, volume.size)
            return {'provider_location': fcd_loc.provider_location()}
        else:
            return self._create_volume_from_fcd(snapshot.provider_location,
                                                snapshot.volume.size, volume)

    def create_cloned_volume(self, volume, src_vref):
        """Creates a clone of the specified volume.

        :param volume: New Volume object
        :param src_vref: Source Volume object
        """
        return self._create_volume_from_fcd(
            src_vref.provider_location, src_vref.size, volume)

    def retype(self, context, volume, new_type, diff, host):
        if not self._storage_policy_enabled:
            return True

        profile = self._get_storage_profile(volume)
        new_profile = self._get_extra_spec_storage_profile(new_type['id'])
        if profile == new_profile:
            LOG.debug("Storage profile matches between new type and old type.")
            return True

        if self._in_use(volume):
            LOG.warning("Cannot change storage profile of attached FCD.")
            return False

        fcd_loc = vops.FcdLocation.from_provider_location(
            volume.provider_location)
        new_profile_id = self.ds_sel.get_profile_id(new_profile)
        self.volumeops.update_fcd_policy(fcd_loc, new_profile_id.uniqueId)
        return True
